from __future__ import annotations
from typing import List, Tuple, Union, Type, Dict

import copy
import itertools
from _balder.device import Device
from _balder.exceptions import IllegalConnectionTypeError


class Connection:
    """
    This is the basic connection class. On one side it is the common base class for all connection objects. On the other
    side it can also be used to describe a container connection, that contains a list of different connection items.
    """
    #: contains all parent connection information for every existing connection
    __parents: Dict[Type[Connection], Dict[str, List[Type[Connection]]]] = {}

    def __init__(self, from_device: Union[Type[Device], None] = None, to_device: Union[Type[Device], None] = None,
                 from_device_node_name: Union[str, None] = None, to_device_node_name: Union[str, None] = None):
        """
        This method creates a new Connection-Tree.

        .. note::
            With a direct instance of a :class:`Connection` object you can create an own Connection-Tree. You can use
            this container object, if you need a container for a list of :class:`Connection`-Trees that are combined
            with an AND (tuple) or/and an OR (list).

        :param from_device: the device this connection starts from (default: None)

        :param to_device: the device this connection ends in (default: None)

        :param from_device_node_name: the node name of the device the connection start

        :param to_device_node_name: the node name of the device the connection ends
        """

        # contains all metadata of this connection object
        self._metadata = {
            "from_device": None, "from_device_node_name": None, "to_device": None, "to_device_node_name": None
        }

        self._bidirectional = None
        # contains all sub connection objects, this connection tree is based on
        self._based_on_connections = []

        if from_device is not None and not issubclass(from_device, Device):
            raise TypeError(f"detect illegal argument element {str(from_device)} for given attribute "
                            f"`from_device` - should be a subclasses of `balder.Device`")
        self._metadata["from_device"] = from_device

        if from_device_node_name is not None and not isinstance(from_device_node_name, str):
            raise TypeError(f"detect illegal argument type {type(from_device_node_name)} for given attribute "
                            f"`from_device_node_name` - should be a string value")
        self._metadata["from_device_node_name"] = from_device_node_name

        if to_device is not None and not issubclass(to_device, Device):
            raise TypeError(f"detect illegal argument element {str(to_device)} for given attribute "
                            f"`to_device` - should be a subclasses of `balder.Device`")
        self._metadata["to_device"] = to_device

        if to_device_node_name is not None and not isinstance(to_device_node_name, str):
            raise TypeError(f"detect illegal argument type {type(to_device_node_name)} for given attribute "
                            f"`to_device_node_name` - should be a string value")
        self._metadata["to_device_node_name"] = to_device_node_name

        # describes if the connection is uni or bidirectional
        # note: currently every connection is bidirectional (we want to add support for this later)
        self._bidirectional = True

        if not ((from_device is None and to_device is None and from_device_node_name is None
                 and to_device_node_name is None) or (
                from_device is not None and to_device is not None and from_device_node_name is not None and
                to_device_node_name is not None)):
            raise ValueError(
                "you have to provide all or none of the following items: `from_device`, `from_device_node_name`, "
                "`to_device` or `to_device_node_name`")

    def __eq__(self, other):
        if isinstance(other, Connection):
            return self.equal_with(other)

        return False

    def __hash__(self):
        all_hashes = hash(self.from_device) + hash(self.to_device) + hash(self.from_node_name) + \
                     hash(self.to_node_name) + hash(str(self))
        for cur_child in self.based_on_elements:
            all_hashes += hash(cur_child)
        return hash(all_hashes)

    def clone(self) -> Connection:
        """
        This method returns an exact clone of this connection. For this clone every inner connection object will be
        newly instantiated, but all internal references (like the devices and so on) will not be copied (objects are the
        same for this object and the clone). The method will make a normal copy for every connection object in the
        `_based_on_elements` list.
        """
        self_copy = copy.copy(self)
        self_copy._based_on_connections = []

        for cur_based_on in self._based_on_connections:
            if isinstance(cur_based_on, tuple):
                cloned_tuple = ()
                for cur_tuple_element in cur_based_on:
                    cloned_tuple = cloned_tuple + (cur_tuple_element.clone(), )
                self_copy._based_on_connections.append(cloned_tuple)
            elif isinstance(cur_based_on, Connection):
                cloned_cur_based_on = cur_based_on.clone()
                self_copy._based_on_connections.append(cloned_cur_based_on)
            else:
                raise TypeError('based on element is not from valid type')
        return self_copy

    @staticmethod
    def __cut_conn_from_only_parent_to_child(elem: Connection) -> List[Connection]:
        """
        This helper method returns all possible pieces while the base element will remain intact - so every returned
        element is from the same type as `elem` (only the related `based_on_elements` are changed).

        .. note::
            The given element itself will also be added here!

        .. note::
            The given element has to be single!
        """
        all_pieces = [elem.clone()]
        if len(elem.based_on_elements) == 0:
            return all_pieces
        # if the next element is a tuple -> call procedure for tuples and add a copy of this object as child
        if isinstance(elem.based_on_elements[0], tuple):
            # return all possibilities of the tuple while adding the current object as child of the tuple
            for cur_tuple in Connection.__cut_tuple_from_only_parent_to_child(elem.based_on_elements[0]):
                # for this we do not have to use `clone`, because we copy the base object with `copy.copy` and
                # completely replace the `_based_on_connections` by our own
                copied_conn = copy.copy(elem)
                copied_conn._based_on_connections = [cur_tuple]
                all_pieces.append(copied_conn)
            return all_pieces
        # if this element is the last element with a parent -> copy it and remove the parent, return it
        if len(elem.based_on_elements[0].based_on_elements) == 0:
            new_elem = copy.copy(elem)
            new_elem._based_on_connections = []
            all_pieces.append(new_elem)
            return all_pieces
        # otherwise, the current item has grandparents, so call the method recursively for parents and add a copy of
        # this object as child
        all_possible_parents = Connection.__cut_conn_from_only_parent_to_child(elem.based_on_elements[0])
        for cur_parent in all_possible_parents:
            copied_conn = copy.copy(elem)
            copied_conn._based_on_connections = [cur_parent]
            all_pieces.append(copied_conn)
        return all_pieces

    @staticmethod
    def __cut_tuple_from_only_parent_to_child(elem: Tuple[Connection]) -> List[Tuple[Connection]]:
        """
        This helper method returns all possible pieces while the base elements of the tuple will remain intact - so
        every returned tuple element is from same type like the related elements in the tuple `elem` (only the related
        `based_on_elements` are changed).

        .. note::
            The given element has to be single!

        .. note::
            Note that this method also returns every possible ordering
        """

        tuple_with_all_possibilities = ()
        for cur_tuple_item in elem:
            tuple_with_all_possibilities += (Connection.__cut_conn_from_only_parent_to_child(cur_tuple_item), )

        cloned_tuple_list = []
        for cur_tuple in list(itertools.product(*tuple_with_all_possibilities)):
            cloned_tuple: Tuple[Connection] = ()
            for cur_tuple_item in cur_tuple:
                cloned_tuple += (cur_tuple_item.clone(), )
            cloned_tuple_list.append(cloned_tuple)
        return cloned_tuple_list

    @staticmethod
    def check_if_tuple_contained_in_connection(tuple_elem: Tuple[Connection], other_elem: Connection) -> bool:
        """
        This method checks if the tuple given by `tuple_elem` is contained in the `other_elem`. To ensure that a tuple
        element is contained in a connection tree, there has to be another tuple into the `other_elem`, that has the
        same length or is bigger. In addition, there has to exist an order combination where every element of the
        `tuple_elem` is contained in the found tuple in `other_elem`. In this case it doesn't matter where the tuple is
        in `other_elem` (will be converted to single, and tuple will be searched in all BASED_ON elements). If the tuple
        element of `other_elem` has fewer items than our `tuple_elem`, it will be ignored. The method only search for
        a valid existing item in the `other_elem` tuple for every item of the `tuple_elem`.

        :param tuple_elem: the tuple element that should be contained in the `other_elem`

        :param other_elem: the connection object, the given tuple should be contained in
        """
        def tuple_is_contained_in_other(inner_tuple, contained_in_tuple):
            # check if every tuple elem fits in one of `contained_in_tuple` (allow to use a position in
            # `contained_in_tuple` multiple times)
            for cur_tuple_element in inner_tuple:
                found_match_for_this_elem = False
                for cur_contained_in_elem in contained_in_tuple:
                    if cur_tuple_element.contained_in(cur_contained_in_elem, ignore_metadata=True):
                        found_match_for_this_elem = True
                        break
                if not found_match_for_this_elem:
                    return False
            return True

        other_singles = other_elem.get_singles()
        tuple_singles = Connection.convert_tuple_to_singles(tuple_elem)
        for cur_tuple_single in tuple_singles:
            for cur_other_single in other_singles:
                # check if we found a tuple element in the other object -> go the single connection upwards and search
                # for a tuple
                cur_sub_other_single: List[Union[Connection, Tuple[Connection]]] = [cur_other_single]
                while len(cur_sub_other_single) > 0:
                    if isinstance(cur_sub_other_single[0], tuple):
                        # found a tuple -> check if length does match
                        if len(cur_sub_other_single[0]) >= len(cur_tuple_single):
                            # length does match
                            if tuple_is_contained_in_other(
                                    cur_tuple_single, contained_in_tuple=cur_sub_other_single[0]):
                                return True
                        # otherwise, this complete element is not possible - skip this single!
                        break
                    cur_sub_other_single = cur_sub_other_single[0].based_on_elements
        return False

    @staticmethod
    def convert_tuple_to_singles(tuple_elem: Tuple[Connection]) -> List[Union[Connection, Tuple[Connection]]]:
        """
        This method converts the given `tuple_elem` to single items and return these.

        :param tuple_elem: the tuple element out of which the single items are being created

        :returns: a list of new connection objects that are single items
        """
        singles_tuple = ()
        tuple_idx = 0
        for cur_tuple_elem in tuple_elem:
            if not isinstance(cur_tuple_elem, Connection):
                raise TypeError(f"the tuple element at index {tuple_idx} of element from `other_conn` is not "
                                "from type `Connection`")
            # get all singles of this tuple element
            singles_tuple += (cur_tuple_elem.get_singles(),)
            tuple_idx += 1
        # now get the variations and add them to our results
        return list(itertools.product(*singles_tuple))

    @staticmethod
    def cleanup_connection_list(full_list: List[Union[Connection, Tuple[Connection]]]) \
            -> List[Union[Connection, Tuple[Connection]]]:
        """
        This method cleanup a connection list while removing items that are direct duplicates and by removing duplicates
        that are fully contained_in other items.

        :param full_list: the full list of connections and tuples of connections that should be cleaned-up

        :returns: returns the cleaned up list
        """
        result = full_list.copy()
        next_loop = True
        while not next_loop:
            next_loop = False
            for cur_elem in result:
                all_other_elems = [cur_item for cur_item in result if cur_item != cur_elem]
                if isinstance(cur_elem, Connection):
                    for cur_other_elem in all_other_elems:
                        # check if it contained in or the same
                        if cur_elem.contained_in(cur_other_elem, ignore_metadata=True):
                            # we can remove it from result list
                            result.remove(cur_elem)
                            next_loop = True
                            break
        return result

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    @classmethod
    def get_parents(cls, tree_name: Union[str, None] = None) -> List[Type[Connection]]:
        """
        This method returns the parent classes of this connection.

        :param tree_name: the tree name the parents should be returned (default: use tree defined in `GlobalSetting`)
        """
        from _balder.balder_session import BalderSession
        if tree_name is None:
            tree_name = BalderSession.get_current_active_global_conntree_name()
        conn_dict = Connection.__parents.get(cls, {})
        parent_list = conn_dict.get(tree_name)
        return [] if parent_list is None else parent_list

    @classmethod
    def set_parents(cls, data: Union[List[Type[Connection]], None], tree_name: str = ""):
        """
        This method allows to set the data for the parent classes of this connection - note that None will remove the
        entry and completely disconnect this object from the global connection tree

        :param data: the new parents (None if everything should be reset)

        :param tree_name: the tree name of the parents that should be set (default: global tree)
        """
        if data is None:
            if cls in Connection.__parents.keys():
                if tree_name in Connection.__parents[cls].keys():
                    del Connection.__parents[cls][tree_name]
        else:
            if Connection.__parents.get(cls) is None:
                Connection.__parents[cls] = {}
            Connection.__parents[cls][tree_name] = data

    @classmethod
    def is_child_of(cls: Type[Connection], other_conn: Type[Connection]) -> bool:
        """
        determines whether this connection is a child of the given connection
        """
        return other_conn.is_parent_of(cls)

    @classmethod
    def is_parent_of(cls: Type[Connection], other_conn: Type[Connection]) -> bool:
        """
        determines whether this connection is a parent of the given connection
        """
        if other_conn.get_parents():
            if cls in other_conn.get_parents():
                return True
            for cur_higher_other_child in other_conn.get_parents():
                # ignore tuples, because this will return False (check with tuples not possible)
                if not isinstance(cur_higher_other_child, tuple):
                    if cls.is_parent_of(other_conn=cur_higher_other_child):
                        return True
            return False
        # the other connection has no parents, so this can not be the parent class
        return False

    @classmethod
    def based_on(cls, *args: Union[Tuple[Union[Type[Connection], Connection], ...], Type[Connection], Connection]) \
            -> Connection:
        """
        With this method it is possible to define several sublayers of the connection. You can pass various other
        connections in this method as arguments.

        Note that multiple parameters of this method (but also of all other methods that work with Connections) mean
        an OR operation. So if you define a `BaseConnType.based_on(ConnType1, ConnType2)` it means that your connection
        `BaseConnType` is based on a `ConnType1` or a `ConnType2`.

        In order to describe an AND operation, i.e. that the `BaseConnType` is based on a` ConnType1` and a `ConnType2`,
        you have to pass them as a tuple. This would look like this: `BaseConnType.based_on((ConnType1, ConnType2))`

        :param args: all connection items for which this connection is based on

        """
        this_instance = cls()

        new_items = []
        for cur_item in args:
            if isinstance(cur_item, Connection):
                if cur_item.__class__ == Connection:
                    # it is a container -> add elements
                    new_items += cur_item.based_on_elements
                else:
                    new_items.append(cur_item)
            elif isinstance(cur_item, type) and issubclass(cur_item, Connection):
                new_items.append(cur_item())
            else:
                new_items.append(cur_item)
        # do not create a container connection if no container is required here
        if cls == Connection and len(new_items) == 1 and not isinstance(new_items[0], tuple):
            return new_items[0]
        this_instance.append_to_based_on(*new_items)
        return this_instance

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def metadata(self) -> dict:
        """returns the connection metadata dictionary"""
        return self._metadata

    @metadata.setter
    def metadata(self, data):
        empty_metadata = {
            "from_device": None, "to_device": None, "from_device_node_name": None, "to_device_node_name": None}

        if not isinstance(data, dict):
            raise ValueError("the given metadata value has to be a dictionary")
        if data != {}:
            if sorted(["from_device", "to_device", "from_device_node_name", "to_device_node_name"]) != \
                    sorted(list(data.keys())):
                raise ValueError("if you provide a metadata dictionary you have to provide all elements of it")
        else:
            data = empty_metadata.copy()

        # only allow to set the metadata dictionary if the old one has the same values or was empty before (no values)
        if data != empty_metadata:
            if self._metadata == empty_metadata:
                # it is ok, because the dictionary was empty before
                pass
            elif self._metadata == data:
                # it is ok too, because the new set data is the same as the data was before
                pass
            else:
                raise ValueError("you can not set another metadata than the data set before - please reset it first")

        self._metadata = data

    @property
    def from_device(self):
        """device from which the connection starts"""
        return self._metadata["from_device"]

    @property
    def to_device(self):
        """device at which the connection ends"""
        return self._metadata["to_device"]

    @property
    def from_node_name(self):
        """the name of the node in the `Device` from which the connection starts"""
        return self._metadata["from_device_node_name"]

    @property
    def to_node_name(self):
        """the name of the node in the `Device` at which the connection ends"""
        return self._metadata["to_device_node_name"]

    @property
    def based_on_elements(self) -> List[Union[Connection, Tuple[Connection]]]:
        """returns a copy of the internal based_on_connection"""
        return self._based_on_connections.copy()

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _metadata_contained_in(self, of_connection: Connection) -> bool:
        """
        This method returns true if the metadata of the current connection is contained in the given one.

        The method returns true in the following situations:
        * both connections are bidirectional and the from and to elements (device and node name) are the same
        * both connections are unidirectional and have the same from and to elements
        * both connections are bidirectional and the from is the to and the to is the from
        * one connection is unidirectional and the other is bidirectional and the from and to elements are the same
        * one connection is unidirectional and the other is bidirectional and the from is the to and the to is the from

        :return: true if the metadata of the current connection is contained in the metadata of the given one
        """
        check_same = \
            self.from_device == of_connection.from_device and self.from_node_name == of_connection.from_node_name and \
            self.to_device == of_connection.to_device and self.to_node_name == of_connection.to_node_name
        check_opposite = \
            self.from_device == of_connection.to_device and self.from_node_name == of_connection.to_node_name and \
            self.to_device == of_connection.from_device and self.to_node_name == of_connection.from_node_name
        if self.is_bidirectional() and of_connection.is_bidirectional():
            return check_same or check_opposite

        if self.is_bidirectional() and not of_connection.is_bidirectional() or \
                not self.is_bidirectional() and of_connection.is_bidirectional():
            return check_same or check_opposite

        return check_same

    def _metadata_equal_with(self, of_connection: Connection) -> bool:
        """
        This method returns true if the metadata of the current connection is equal with the metadata of the given
        connection.

        The method returns true in the following situations:
        * both connections are bidirectional and the from and to elements (device and node name) are the same
        * both connections are unidirectional and have the same from and to elements
        * both connections are bidirectional and the from is the to and the to is the from

        :return: true if the metadata of the current connection is contained in the metadata of the given one
        """
        check_same = \
            self.from_device == of_connection.from_device and self.from_node_name == of_connection.from_node_name and \
            self.to_device == of_connection.to_device and self.to_node_name == of_connection.to_node_name
        check_opposite = \
            self.from_device == of_connection.to_device and self.from_node_name == of_connection.to_node_name and \
            self.to_device == of_connection.from_device and self.to_node_name == of_connection.from_node_name
        if self.is_bidirectional() and of_connection.is_bidirectional() or \
                not self.is_bidirectional() and not of_connection.is_bidirectional():
            return check_same or check_opposite

        return False

    def _get_intersection_with_other_single(self, other_conn: Union[Connection, Tuple[Connection]]) \
            -> List[Connection, Tuple[Connection]]:
        """
        A helper method that returns an intersection between the two connections (self and the given one).

        :param other_conn: the other **single** connection object (could be a tuple too, but note that this has to have
                           only **single** elements!)
        """
        if not self.is_single():
            raise ValueError(
                "the current connection object is not single -> method only possible for single connections")
        if isinstance(other_conn, Connection):
            if other_conn.__class__ == Connection:
                raise ValueError("a container object from direct class `Connection` is not allowed here - please "
                                 "provide a single connection or a single tuple")
            if not other_conn.is_single():
                raise ValueError(
                    "the connection object given by `other_conn` is not single -> method only possible for single "
                    "connections")
        elif isinstance(other_conn, tuple):
            # has to be a tuple
            cur_idx = 0
            for cur_tuple_element in other_conn:
                if not cur_tuple_element.is_single():
                    raise ValueError(
                        f"the connection object given by tuple element at index {cur_idx} in `other_conn` is not "
                        f"single -> method only possible for single connections")
                cur_idx += 1
            other_conn = Connection.based_on(other_conn)
        else:
            raise TypeError("the object given by `other_conn` has to be from type `Connection` or has to be a tuple "
                            "of `Connection` objects")

        intersection = []

        #: check if some sub elements of self connection are contained in `other_conn`
        self_pieces = self.cut_into_all_possible_subtrees()
        for cur_piece in self_pieces:
            if isinstance(cur_piece, Connection):
                if cur_piece.contained_in(other_conn, ignore_metadata=True):
                    intersection.append(cur_piece)
            else:
                # isinstance of tuple
                if Connection.check_if_tuple_contained_in_connection(cur_piece, other_conn):
                    intersection.append(cur_piece)

        #: check if some sub elements of `other_conn` are contained in self connection
        other_pieces = other_conn.cut_into_all_possible_subtrees()
        for cur_piece in other_pieces:
            if isinstance(cur_piece, Connection):
                if cur_piece.contained_in(self, ignore_metadata=True):
                    intersection.append(cur_piece)
            else:
                # isinstance of tuple
                if Connection.check_if_tuple_contained_in_connection(cur_piece, self):
                    intersection.append(cur_piece)

        #: filter all duplicated (and contained in each other) connections
        intersection_without_duplicates = []
        for cur_conn in intersection:
            checkable_cur_conn = Connection.based_on(cur_conn) if isinstance(cur_conn, tuple) else cur_conn
            found_it = False
            for cur_existing_conn in intersection_without_duplicates:
                checkable_cur_existing_conn = Connection.based_on(cur_existing_conn) \
                    if isinstance(cur_existing_conn, tuple) else cur_existing_conn
                if checkable_cur_conn.equal_with(checkable_cur_existing_conn, ignore_metadata=True):
                    found_it = True
                    break
            if not found_it:
                intersection_without_duplicates.append(cur_conn)
        intersection_filtered = []
        #: filter all *contained in each other* connections
        for cur_conn in intersection_without_duplicates:
            usable_cur_conn = Connection.based_on(cur_conn) if isinstance(cur_conn, tuple) else cur_conn
            is_contained_in_another = False
            for cur_validate_cnn in intersection_without_duplicates:
                cur_usable_validate_cnn = \
                    Connection.based_on(cur_validate_cnn) if isinstance(cur_validate_cnn, tuple) else cur_validate_cnn
                if cur_validate_cnn == cur_conn:
                    # skip the same element
                    continue
                if usable_cur_conn.contained_in(cur_usable_validate_cnn, ignore_metadata=True):
                    is_contained_in_another = True
                    break
            if not is_contained_in_another:
                intersection_filtered.append(cur_conn)

        # only return unique objects
        return intersection_filtered

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def set_devices(self, from_device: Type[Device], to_device: Type[Device]):
        """
        Method for setting the devices of the connection if this has not yet been done during instantiation

        .. note::
            Note that you can not change the devices after they were set (only possible internally)

        :param from_device: device from which the connection starts

        :param to_device: device at which the connection ends
        """
        if self.from_device is not None or self.to_device is not None:
            raise ValueError("devices already set")
        self._metadata["from_device"] = from_device
        self._metadata["to_device"] = to_device

    def update_node_names(self, from_device_node_name: str, to_device_node_name: str) -> None:
        """
        This method dates the names of the nodes from which the connection in the `from_device` originates and finally
        arrives in the` to_device`. Please provide the node name of your ``from_device`` in ``from_device_node_name``
        and the node name of your ``to_device`` in ``to_device_node_name``.

        :param from_device_node_name: specifies the node in the ``from_device``

        :param to_device_node_name: specifies the node in the ``to_device``
        """
        self._metadata["from_device_node_name"] = from_device_node_name
        self._metadata["to_device_node_name"] = to_device_node_name

    def cut_into_all_possible_subtrees(self) -> List[Union[Connection, Tuple[Connection]]]:
        """
        This method cuts the resolved connection tree in all possible pieces by removing elements that change
        the existing tree - thereby the method returns a list with all possibilities (a copy of this object with all
        elements is contained too).

        .. note::
            Note that the given element or the child elements of a given direct :class:`Connection` object has to be
            single!
        """
        if self.__class__ == Connection:
            # this is only a container, execute process for every item of this one
            child_elems = self.based_on_elements
        else:
            child_elems = [self]

        all_pieces = []

        for cur_item in child_elems:

            if isinstance(cur_item, Connection):
                if not cur_item.is_single():
                    raise ValueError("one of the given element is not single -> method only works with single items")
            elif isinstance(cur_item, tuple):
                tuple_idx = 0
                for cur_tuple_elem in cur_item:
                    if not cur_tuple_elem.is_single():
                        raise ValueError(
                            f"one of the given tuple element has a item at index {tuple_idx}, that is not single -> "
                            f"method only works with single items or tuple of single items")
                    tuple_idx += 1

            cur_element = [cur_item]
            while len(cur_element) > 0:
                if isinstance(cur_element[0], Connection):
                    all_pieces += Connection.__cut_conn_from_only_parent_to_child(cur_element[0])
                elif isinstance(cur_element[0], tuple):
                    all_pieces += Connection.__cut_tuple_from_only_parent_to_child(cur_element[0])
                    # now we can not go deeper, because we need this current AND connection
                    break
                cur_element = cur_element[0].based_on_elements

        # filter all duplicates
        return list(set(all_pieces))

    def set_metadata_for_all_subitems(self, metadata: Union[Dict[str, Union[Device, str]], None]):
        """
        This method sets the metadata for all existing Connection items in this element.

        :param metadata: the metadata that should be set (if the value is explicitly set to None, it removes the
                         metadata from every item)
        """
        if metadata is None:
            metadata = {
                "from_device": None, "to_device": None, "from_device_node_name": None, "to_device_node_name": None}

        for cur_base_elem in self._based_on_connections:
            if isinstance(cur_base_elem, Connection):
                cur_base_elem.set_metadata_for_all_subitems(metadata=metadata)
            elif isinstance(cur_base_elem, tuple):
                for cur_tuple_elem in cur_base_elem:
                    cur_tuple_elem.set_metadata_for_all_subitems(metadata=metadata)
            else:
                raise TypeError(
                    f"found illegal type {cur_base_elem.__class__} for based_on item at index {cur_base_elem}")
        self.metadata = metadata

    def get_tree_str(self) -> str:
        """
        This method returns a string, that represents all internal connections as readable string.

        :return: a readable string of the whole connection tree
        """
        if len(self.based_on_elements) == 0:
            return f"{self.__class__.__name__}()"

        based_on_strings = []
        for cur_elem in self.based_on_elements:
            if isinstance(cur_elem, tuple):
                based_on_strings.append(
                    f"({', '.join([cur_tuple_elem.get_tree_str() for cur_tuple_elem in cur_elem])})")
            else:
                based_on_strings.append(cur_elem.get_tree_str())

        return f"{self.__class__.__name__}.based_on({', '.join(based_on_strings)})"

    def is_bidirectional(self):
        """
        Provides the information in which direction the connection is supported (if the property returns true, the
        connection must work in both directions, otherwise it is a unidirectional connection from the `from_device` to
        the` to_device`
        """
        return self._bidirectional

    def is_universal(self):
        """
        Provides the information if the current connection object is a universal connection. This means, that the type
        is the base :class:`Connection` and the based_on_elements are empty.
        """
        return self.__class__ == Connection and len(self._based_on_connections) == 0

    def is_resolved(self):
        """
        This method returns true if the given (sub-)connection tree is completely resolved (describes every single
        Connection object of the tree - there are no undefined connection layers between an object and the given
        parent).
        """
        for cur_based_on in self.based_on_elements:
            if isinstance(cur_based_on, tuple):
                for cur_tuple_elem in cur_based_on:
                    # check the next parent class type only if the main self type is not a container
                    # (base `balder.Connection` object)
                    if self.__class__ != Connection:
                        if cur_tuple_elem not in self.__class__.get_parents():
                            # only one element is not directly parent -> return false
                            return False
                    if not cur_tuple_elem.is_resolved():
                        # only one subelement is not resolved -> return false
                        return False
            else:
                # no tuple, single element

                # check the next parent class type only if the main self type is not a container
                # (base `balder.Connection` object)
                if self.__class__ != Connection:
                    if cur_based_on.__class__ not in self.__class__.get_parents():
                        # only one element is not directly parent -> return false
                        return False
                if not cur_based_on.is_resolved():
                    # only one subelement is not resolved -> return false
                    return False
        return True

    def is_single(self):
        """
        This method returns true if there exists no logical **OR** in the based on connection(s). One **AND** tuple is
        allowed.

        .. note::
            Note that this method also returns False, if the connection is not completely resolved!
        """
        if len(self.based_on_elements) == 0:
            # tree ends here
            return True

        if len(self.based_on_elements) > 1:
            # more than one element -> not single
            return False

        if not self.is_resolved():
            return False

        if isinstance(self.based_on_elements[0], tuple):
            for cur_tuple_elem in self.based_on_elements[0]:
                # if only one element of tuple is not single -> return False
                if not cur_tuple_elem.is_single():
                    return False
            # all elements are single
            return True

        return self.based_on_elements[0].is_single()

    def get_resolved(self) -> Connection:
        """
        This method returns a resolved Connection Tree. This means that it convert the based_on references so that
        every based on connection is a direct parent of the current element. It secures that there are no undefined
        connection layers between an object and the given parent.

        .. note::
            This method returns the connection without a container :class:`Connection`, if the container
            :class:`Connection` would only consist of one single connection (which is no tuple!). In this case the
            method returns this child connection directly without any :class:`Connection` container otherwise the
            :class:`Connection` container class with all resolved child classes will be returned.
        """
        copied_base = copy.copy(self)
        if not copied_base.is_resolved():
            copied_base._based_on_connections = []
            if self.__class__ == Connection:
                # the base object is a container Connection - iterate over the items and determine the values for them
                for cur_item in self.based_on_elements.copy():
                    if isinstance(cur_item, tuple):
                        new_tuple = ()
                        for cur_tuple_item in cur_item:
                            new_tuple += (cur_tuple_item.get_resolved(), )
                        copied_base._based_on_connections.append(new_tuple)
                    else:
                        copied_base._based_on_connections.append(cur_item.get_resolved())
            else:
                for next_higher_parent in self.based_on_elements:
                    if isinstance(next_higher_parent, tuple):
                        # determine all possibilities
                        direct_ancestors_tuple = ()
                        for cur_tuple_elem in next_higher_parent:
                            if cur_tuple_elem.__class__ in self.__class__.get_parents():
                                # element already is a direct ancestor
                                direct_ancestors_tuple += (cur_tuple_elem, )
                            else:
                                all_pos_possibilities = []
                                # add all possible direct parents to the possibilities list
                                for cur_direct_parent in self.__class__.get_parents():
                                    if cur_direct_parent.is_parent_of(cur_tuple_elem.__class__):
                                        all_pos_possibilities.append(cur_direct_parent.based_on(cur_tuple_elem))
                                direct_ancestors_tuple += (all_pos_possibilities, )
                        # resolve the opportunities and create multiple possible tuples where all elements are direct
                        # parents
                        for cur_possibility in itertools.product(*direct_ancestors_tuple):
                            new_child_tuple = ()
                            for cur_tuple_element in cur_possibility:
                                new_child_tuple += (cur_tuple_element.get_resolved(), )
                            copied_base._based_on_connections.append(new_child_tuple)
                    else:
                        if next_higher_parent.__class__ in self.__class__.get_parents():
                            # is already a direct parent
                            copied_base._based_on_connections.append(next_higher_parent.get_resolved())
                        else:
                            # only add the first level of direct parents - deeper will be added by recursively call of
                            # `get_resolved`
                            for cur_self_direct_parent in self.__class__.get_parents():
                                if next_higher_parent.__class__.is_parent_of(cur_self_direct_parent):
                                    new_child = cur_self_direct_parent.based_on(next_higher_parent)
                                    copied_base._based_on_connections.append(new_child.get_resolved())
        # if it is a connection container, where only one element exists that is no tuple -> return this directly
        # instead of the container
        if copied_base.__class__ == Connection and len(copied_base.based_on_elements) == 1 and not \
                isinstance(copied_base.based_on_elements[0], tuple):
            return copied_base.based_on_elements[0]

        return copied_base

    def get_singles(self) -> List[Connection]:
        """
        This method returns a list with connection objects where every element is a single clear possibility of the
        current connection sub-tree. With this method the **OR** possibilities will be resolved, so that every item does
        not contain any OR relations.

        .. note::
            If the current object is a container :class:`Connection` object (direct instance of :class:`Connection`),
            this method returns the single elements without a container class!
        """
        all_singles = []

        all_self_items = [self]
        if self.__class__ == Connection and len(self.based_on_elements) == 0:
            return [self]
        if self.__class__ == Connection:
            all_self_items = self.based_on_elements
        for cur_item in all_self_items:
            if isinstance(cur_item, tuple):
                all_singles += Connection.convert_tuple_to_singles(cur_item)
            elif cur_item.is_single():
                all_singles.append(copy.copy(cur_item))
            else:
                # do this for resolved version of this object
                resolved_obj = cur_item.get_resolved()

                if len(resolved_obj.based_on_elements) == 0:
                    # element has no children -> return only this
                    return [resolved_obj]
                for cur_child in resolved_obj.based_on_elements:
                    for cur_single_child in cur_child.get_singles():
                        copied_base = copy.copy(resolved_obj)
                        copied_base._based_on_connections = [cur_single_child]
                        all_singles.append(copied_base)
        # convert all tuple objects in a connection object and set metadata of self object
        cleaned_singles = []

        for cur_single in all_singles:
            if isinstance(cur_single, Connection):
                new_cnn = cur_single
            else:
                new_cnn = Connection.based_on(cur_single)

            new_cnn.set_metadata_for_all_subitems(self.metadata)
            cleaned_singles.append(new_cnn)
        return cleaned_singles

    def get_conn_partner_of(self, device: Type[Device], node: Union[str, None] = None) -> Tuple[Type[Device], str]:
        """
        This method returns the connection partner of this connection - it always returns the other not given side

        :param device: the device itself - the other will be returned

        :param node: the node name of the device itself (only required if the connection starts and ends with the same
                     device)
        """
        if device not in (self.from_device, self.to_device):
            raise ValueError(f"the given device `{device.__qualname__}` is no component of this connection")
        if node is None:
            # check that the from_device and to_device are not the same
            if self.from_device == self.to_device:
                raise ValueError("the connection is a inner-device connection (start and end is the same device) - you "
                                 "have to provide the `node` string too")
            if device == self.from_device:
                return self.to_device, self.to_node_name

            return self.from_device, self.from_node_name

        if node not in (self.from_node_name, self.to_node_name):
            raise ValueError(f"the given node `{node}` is no component of this connection")

        if device == self.from_device and node == self.from_node_name:
            return self.to_device, self.to_node_name

        if device == self.to_device and node == self.to_node_name:
            return self.from_device, self.from_node_name

        raise ValueError(f"the given node `{node}` is no component of the given device `{device.__qualname__}`")

    def has_connection_from_to(self, start_device, end_device=None):
        """
        This method checks if there is a connection from ``start_device`` to ``end_device``. This will return
        true if the ``start_device`` and ``end_device`` given in this method are also the ``start_device`` and
        ``end_device`` mentioned in this connection object. If this is a bidirectional connection, ``start_device`` and
        ``end_device`` can switch places.


        :param start_device: the device for which the method should check whether it is a communication partner (for
                             non-bidirectional connection, this has to be the start device)

        :param end_device: the other device for which the method should check whether it is a communication partner (for
                           non-bidirectional connection, this has to be the end device - this is optional if only the
                           start device should be checked)

        :return: returns true if the given direction is possible
        """
        if end_device is None:

            if self.is_bidirectional():
                return start_device in (self.from_device, self.to_device)

            return start_device == self.from_device

        if self.is_bidirectional():
            return start_device == self.from_device and end_device == self.to_device or \
                   start_device == self.to_device and end_device == self.from_device

        return start_device == self.from_device and end_device == self.to_device

    def equal_with(self, other_conn: Connection, ignore_metadata=False) -> bool:
        """
        This method returns True if the current object matches the given object. It always converts the elements to a
        resolved version and checks if both of them are exactly the same.

        .. note::
            Note that both Connection objects have to have the same ending parent elements. Only the order of the parent
            elements are obviously irrelevant.

        .. note::
            Note that it doesn't matter if the connection is embedded in a container-Connection element (direct
            instance of :class`Connection`) or not. It only checks, that the logical data of them are the same. If both
            elements are a container for a list of child connections, the method secures that both has the same
            children. If only one (same for both) has one child connection which is embedded in a container
            :class:`Connection` class, it returns also true if the other connection is like the one child element of
            the other container :class:`Connection`.

        :param other_conn: the other connection, this connection will be compared with

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: returns True if both elements are same
        """
        if self.__class__ not in (Connection, other_conn.__class__):
            return False

        if not ignore_metadata:
            metadata_check_result = self._metadata_equal_with(of_connection=other_conn)
            if metadata_check_result is False:
                return False

        # we do not need to check something for the container :class:`Connection`, because the `.get_resolved()` will
        # return no container if it only consists of one child connection
        resolved_self = self.get_resolved()
        resolved_other = other_conn.get_resolved()

        def check_elems_if(elems_from, are_in_elems_from):
            for cur_elem in elems_from:
                found_equal = False
                for cur_other_elem in are_in_elems_from:
                    if cur_elem.equal_with(cur_other_elem, ignore_metadata=ignore_metadata):
                        found_equal = True
                        break
                if not found_equal:
                    return False
            return True

        def check_tuples_if(tuples_from, are_in_tuples_from):
            for cur_search_tuple in tuples_from:
                found_match_for_cur_search_tuple = False
                # go through each unmatched other tuple
                for cur_other_tuple in are_in_tuples_from:
                    cur_search_tuple_is_completely_in_other_tuple = True
                    for cur_search_tuple_elem in cur_search_tuple:
                        fount_it = False
                        for cur_other_tuple_elem in cur_other_tuple:
                            if cur_search_tuple_elem.equal_with(cur_other_tuple_elem, ignore_metadata=ignore_metadata):
                                fount_it = True
                                break
                        if not fount_it:
                            cur_search_tuple_is_completely_in_other_tuple = False
                            break
                    if cur_search_tuple_is_completely_in_other_tuple:
                        # here we have no match
                        found_match_for_cur_search_tuple = True
                        break
                if not found_match_for_cur_search_tuple:
                    return False
            return True

        self_based_on_elems = [
            cur_elem for cur_elem in resolved_self.based_on_elements if isinstance(cur_elem, Connection)]
        self_based_on_tuples = [
            cur_elem for cur_elem in resolved_self.based_on_elements if isinstance(cur_elem, tuple)]
        other_based_on_elems = [
            cur_elem for cur_elem in resolved_other.based_on_elements if isinstance(cur_elem, Connection)]
        other_based_on_tuples = [
            cur_elem for cur_elem in resolved_other.based_on_elements if isinstance(cur_elem, tuple)]

        # check single connection elements (if they match all in both directions)
        if not check_elems_if(elems_from=self_based_on_elems, are_in_elems_from=other_based_on_elems) or \
                not check_elems_if(elems_from=other_based_on_elems, are_in_elems_from=self_based_on_elems):
            return False

        # check tuple connection elements (if they match all in both directions)
        if not check_tuples_if(tuples_from=self_based_on_tuples, are_in_tuples_from=other_based_on_tuples) or \
                not check_tuples_if(tuples_from=other_based_on_tuples, are_in_tuples_from=self_based_on_tuples):
            return False

        return True

    def contained_in(self, other_conn: Connection, ignore_metadata=False) -> bool:
        """
        This method helps to find out whether this connection-tree fits within a given connection tree. A connection
        object is a certain part of the large connection tree that Balder has at its disposal. This method checks
        whether a possibility of this connection tree fits in one possibility of the given connection tree.

        .. note::
            The method returns true if one single connection of this object fits in another single connection that is
            given by `other_conn`.

        :param other_conn: the other connection

        :param ignore_metadata: if this value is true the method ignores the metadata

        :return: true if the self object is contained in the `other_conn`, otherwise false
        """
        if not ignore_metadata:
            metadata_check_result = self._metadata_contained_in(of_connection=other_conn)
            if metadata_check_result is False:
                return False

        resolved_self = self.get_resolved()
        resolved_other = other_conn.get_resolved()

        if resolved_self.__class__ == Connection and len(resolved_self.based_on_elements) == 0 or \
                resolved_other.__class__ == Connection and len(resolved_other.based_on_elements) == 0:
            # one of the resolved object is a raw `Connection` object without based-on-elements -> always true
            return True

        self_possibilities = [resolved_self]
        if resolved_self.__class__ == Connection:
            self_possibilities = resolved_self.based_on_elements
        for cur_self in self_possibilities:
            if isinstance(cur_self, tuple):
                if Connection.check_if_tuple_contained_in_connection(cur_self, resolved_other):
                    return True
            elif not cur_self.__class__ == resolved_other.__class__:
                for cur_based_on_elem in resolved_other.based_on_elements:
                    if isinstance(cur_based_on_elem, tuple):
                        # check if the current connection fits in one of the tuple items -> allowed too (like a smaller
                        # AND contained in a bigger AND)
                        for cur_other_tuple_element in cur_based_on_elem:
                            if cur_self.contained_in(cur_other_tuple_element, ignore_metadata=ignore_metadata):
                                return True
                    else:
                        if cur_self.contained_in(cur_based_on_elem, ignore_metadata=ignore_metadata):
                            # element was found in this branch
                            return True
            else:
                # The element itself has already matched, now we still have to check whether at least one tuple or a
                # connection (i.e. one element of the list) is matched with one of the other
                singles_self = cur_self.get_singles()
                singles_other = resolved_other.get_singles()

                # check that for one single_self element all hierarchical based_on elements are in one of the single
                # other element
                for cur_single_self in singles_self:
                    for cur_single_other in singles_other:
                        # check if both consists of only one element
                        if len(cur_single_self.based_on_elements) == 0:
                            # the cur self single is only one element -> this is contained in the other
                            return True

                        if len(cur_single_other.based_on_elements) == 0:
                            # the other element is only one element, but the self element not -> contained_in
                            # for this single definitely false
                            continue

                        # note: for both only one `based_on_elements` is possible, because they are singles
                        if isinstance(cur_single_self.based_on_elements[0], tuple) and \
                                isinstance(cur_single_other.based_on_elements[0], tuple):
                            # both are tuples -> check that there exist a matching where all are contained_in
                            tuple_is_complete_contained_in = True
                            # check that every tuple element of self is contained in minimum one tuple elements of each
                            # other
                            for cur_self_tuple_element in cur_single_self.based_on_elements[0]:
                                find_some_match_for_cur_self_tuple_element = False
                                for cur_other_tuple_element in cur_single_other.based_on_elements[0]:
                                    if cur_self_tuple_element.contained_in(
                                            cur_other_tuple_element, ignore_metadata=ignore_metadata):
                                        # find a match, where the current tuple element is contained in one tuple
                                        # element of the other
                                        find_some_match_for_cur_self_tuple_element = True
                                        break
                                if not find_some_match_for_cur_self_tuple_element:
                                    tuple_is_complete_contained_in = False
                                    break
                            if tuple_is_complete_contained_in:
                                # find a complete valid match
                                return True
                        elif isinstance(cur_single_self.based_on_elements[0], Connection) and \
                                isinstance(cur_single_other.based_on_elements[0], Connection):
                            # both are connection trees -> check if the subtrees are contained in
                            if cur_single_self.based_on_elements[0].contained_in(
                                    cur_single_other.based_on_elements[0], ignore_metadata=ignore_metadata):
                                # find a complete valid match
                                return True
                        elif isinstance(cur_single_self.based_on_elements[0], tuple) and \
                                isinstance(cur_single_other.based_on_elements[0], Connection):
                            # this is allowed too, if every tuple item is contained_in the relevant connection
                            # check that every tuple element of self is contained in the other connection
                            for cur_self_tuple_element in cur_single_self.based_on_elements[0]:
                                if not cur_self_tuple_element.contained_in(
                                        cur_single_other, ignore_metadata=ignore_metadata):
                                    # find a match, where the current tuple element is not contained in the other
                                    # connection -> tuple can not be contained in
                                    return False
                            return True
                        # skip all others possibilities
        return False

    def intersection_with(
            self, other_conn: Union[Connection, Type[Connection],
                                    List[Connection, Type[Connection], Tuple[Connection]]]) \
            -> Union[Connection, None]:
        """
        This method returns a list of subtrees that describe the intersection of this connection subtree and the
        given ones. Note that this method converts all connections in **single** **resolved** connections first.
        For these connections the method checks if there are common intersections between the elements of this object
        and the given connection elements.
        The method cleans up this list and only return unique sub-connection trees!

        :param other_conn: the other sub connection tree list

        :return: the intersection connection or none, if the method has no intersection
        """
        if isinstance(other_conn, type):
            if not issubclass(other_conn, Connection):
                raise TypeError("the given `other_conn` has to be from type `Connection`")
            other_conn = other_conn()

        if isinstance(other_conn, Connection):
            if other_conn.__class__ == Connection:
                if len(other_conn.based_on_elements) == 0:
                    return self.clone()
                other_conn = other_conn.based_on_elements
            else:
                other_conn = [other_conn]

        if self.__class__ == Connection and len(self.based_on_elements) == 0:
            return Connection.based_on(*other_conn).clone()

        # determine all single connection of the two sides (could contain tuple, where every element is a single
        # connection too)
        self_conn_singles = []
        other_conn_singles = []

        idx = 0
        for cur_self_elem in self.get_singles():
            if isinstance(cur_self_elem, Connection):
                self_conn_singles.append(cur_self_elem)
            elif isinstance(cur_self_elem, type) and issubclass(cur_self_elem, Connection):
                self_conn_singles.append(cur_self_elem())
            else:
                raise TypeError(f"the element at index {idx} of `self_conn` is not from type `Connection` or is a "
                                f"tuple of that")

        idx = 0
        for cur_other_elem in other_conn:
            if isinstance(cur_other_elem, tuple):
                other_conn_singles += Connection.convert_tuple_to_singles(cur_other_elem)
            elif isinstance(cur_other_elem, Connection):
                other_conn_singles += cur_other_elem.get_singles()
            elif isinstance(cur_other_elem, type) and issubclass(cur_other_elem, Connection):
                other_conn_singles.append(cur_other_elem())
            else:
                raise TypeError(f"the element at index {idx} of `other_conn` is not from type `Connection` or is a "
                                f"tuple of that")
            idx += 1

        intersections = []
        # determine intersections between all of these single components
        for cur_self_conn in self_conn_singles:
            for cur_other_conn in other_conn_singles:
                for cur_intersection in cur_self_conn._get_intersection_with_other_single(cur_other_conn):
                    if isinstance(cur_intersection, tuple):
                        cur_intersection = Connection.based_on(cur_intersection)
                    if cur_intersection not in intersections:
                        intersections.append(cur_intersection)

        #: filter all *contained in each other* connections
        intersection_filtered = []
        for cur_conn in intersections:
            is_contained_in_another = False
            for cur_validate_cnn in intersections:
                if cur_validate_cnn == cur_conn:
                    # skip the same element
                    continue
                if cur_conn.contained_in(cur_validate_cnn, ignore_metadata=True):
                    is_contained_in_another = True
                    break
            if not is_contained_in_another:
                intersection_filtered.append(cur_conn)

        if len(intersection_filtered) == 0:
            # there is no intersection
            return None
        if len(intersection_filtered) > 1 or isinstance(intersection_filtered[0], tuple):
            return Connection.based_on(*intersection_filtered).clone()

        return intersection_filtered[0].clone()

    def append_to_based_on(self, *args: Union[Tuple[Union[Type[Connection], Connection]]]) -> None:
        """
        with this method you can extend the internal based_on list with the transferred elements. Any number of
        :meth:`Connection` objects or tuples with :meth:`Connection` objects can be given to this method

        :param args: all connection items that should be added here
        """

        for cur_idx, cur_connection in enumerate(args):
            if isinstance(cur_connection, type):
                if not issubclass(cur_connection, Connection):
                    raise TypeError(f"illegal type `{cur_connection.__name__}` for parameter number {cur_idx}")
                if self.__class__ != Connection:
                    if not cur_connection.is_parent_of(self.__class__):
                        raise IllegalConnectionTypeError(
                            f"the given connection `{cur_connection.__name__}` (parameter pos {cur_idx}) is no parent "
                            f"class of the `{self.__class__.__name__}`")
                # this is a simple Connection type object -> simply add an instance of it to the full list
                new_conn = cur_connection()
                self._based_on_connections.append(new_conn)

            elif isinstance(cur_connection, Connection):
                if cur_connection.__class__ == Connection:
                    raise ValueError(f"it is not allowed to provide a container Connection object in based_on items - "
                                     f"found at index {cur_idx}")
                # `based_on` call for this sub-connection because we get an `Connection` object
                if self.__class__ != Connection:
                    if not cur_connection.__class__.is_parent_of(self.__class__):
                        raise IllegalConnectionTypeError(
                            f"the given connection `{cur_connection.__class__.__name__}` (parameter pos {cur_idx}) is "
                            f"no parent class of the `{self.__class__.__name__}`")
                self._based_on_connections.append(cur_connection)

            elif isinstance(cur_connection, tuple):
                result_tuple = ()
                for cur_tuple_idx, cur_tuple_elem in enumerate(cur_connection):
                    if isinstance(cur_tuple_elem, type):
                        if not issubclass(cur_tuple_elem, Connection):
                            raise TypeError(f"illegal type `{cur_tuple_elem.__name__}` for tuple element "
                                            f"{cur_tuple_idx} for parameter number {cur_idx}")
                        if self.__class__ != Connection:
                            if not cur_tuple_elem.is_parent_of(self.__class__):
                                raise IllegalConnectionTypeError(
                                    f"the given connection `{cur_tuple_elem.__name__}` (tuple element {cur_tuple_idx} "
                                    f"for parameter at pos {cur_idx}) is no parent class of the "
                                    f"`{self.__class__.__name__}`")
                        # this is a simple Connection type object -> simply add the instance of connection to result
                        # tuple
                        result_tuple += (cur_tuple_elem(), )
                    elif isinstance(cur_tuple_elem, Connection):
                        # `based_on` call for this sub-connection because we get an `Connection` object
                        if self.__class__ != Connection:
                            if not cur_tuple_elem.__class__.is_parent_of(self.__class__):
                                raise IllegalConnectionTypeError(
                                    f"the given connection `{cur_tuple_elem.__class__.__name__}` (tuple element "
                                    f"{cur_tuple_idx} for parameter at pos {cur_idx}) is no parent class of the "
                                    f"`{self.__class__.__name__}`")
                        result_tuple += (cur_tuple_elem, )
                    elif isinstance(cur_tuple_elem, tuple):
                        raise TypeError(f"nested tuples (tuple element {cur_tuple_idx} for parameter at pos {cur_idx}) "
                                        f"and thus nested AND operations are not possible")
                    elif isinstance(cur_tuple_elem, list):
                        raise TypeError(f"nested lists (tuple element {cur_tuple_idx} for parameter at pos {cur_idx}) "
                                        f"and thus nested OR/AND operations are not possible")
                    else:
                        raise TypeError(f"illegal type `{cur_tuple_elem.__name__}` for tuple element {cur_tuple_idx} "
                                        f"for parameter at pos {cur_idx}")
                self._based_on_connections.append(result_tuple)
            else:
                raise TypeError(f"illegal type `{cur_connection.__name__}` for parameter at pos {cur_idx}")



from typing import List
import sys
import traceback
from typing import Union
from abc import ABC, abstractmethod
import pathlib
from multiprocessing import Process, Queue
from _balder.balder_session import BalderSession
from _balder.console.balder import _console_balder_debug
from _balder.exceptions import BalderException
from .observer_compare import compare_observed_list_with_expected


class Base0EnvtesterClass(ABC):
    """
    This is a base class for all tests that want to use the 0_envtester environment.
    """

    @property
    @abstractmethod
    def expected_data(self) -> tuple:
        """This property should return the expected data that should be sent by the process over the queue."""

    @property
    def expected_data_alternative(self) -> Union[tuple, None]:
        """
        optional property that can hold an alternative expected data (often required, if it is not clear at which
        branch will be executed first while an error is expected in one of them
        """
        return None

    @property
    def expected_exit_code(self) -> int:
        """returns the expected exit code the process should terminate"""
        return 0

    @property
    def cmd_args(self):
        """returns the command line arguments that should be used in the process call"""
        return []

    def test(self, balder_working_dir):
        """
        This is the general test method, that executes the balder process.
        """
        exec_log_queue = Queue(maxsize=0)

        proc = Process(target=self.__class__.processed_method,
                       args=(balder_working_dir, self.cmd_args, exec_log_queue,
                             self.__class__.validate_finished_session, self.__class__.handle_balder_exception))
        proc.start()
        exec_list = []
        while proc.is_alive() or not exec_log_queue.empty():

            if not exec_log_queue.empty():
                new_data = exec_log_queue.get()
                new_data["file"] = pathlib.Path(new_data["file"]).relative_to(balder_working_dir)
                exec_list.append(new_data)

        proc.join()
        assert proc.exitcode == self.expected_exit_code, \
            f"the process terminates with unexpected exit code `{proc.exitcode}` - exit code " \
            f"`{self.expected_exit_code}` was expected"

        if self.expected_data_alternative is None:
            # only one possible truth
            compare_observed_list_with_expected(exec_list, self.expected_data)
        else:
            # two possibilities are allowed
            error1 = None
            try:
                compare_observed_list_with_expected(exec_list, self.expected_data)
            except Exception as exc:
                error1 = exc
            error2 = None
            try:
                compare_observed_list_with_expected(exec_list, self.expected_data_alternative)
            except Exception as exc:
                error2 = exc
            if error1 and error2:
                # can not find possible matching
                raise error1

    @staticmethod
    def processed_method(env_dir, args: List[str], queue: Queue, cb_validate_finished_session, cb_balder_exception):
        """
        This is the processed method

        :param env_dir: the directory path to the `env` directory for the current test

        :param args: a list with all command line arguments that should be used for the balder call

        :param queue: the queue the :class:RuntimeObserver` uses to communicate with the pytest process

        :param cb_validate_finished_session: a callback that validates the :class:`BalderSession`, after the session
                                             has finished

        :param cb_balder_exception: will be executed in case a balder error was thrown during collecting or resolving
        """
        # use the BalderSession object to import the balderglob file - than load the
        def set_queue(session):
            RuntimeObserver = getattr(sys.modules.get('env.balderglob'), "RuntimeObserver")
            RuntimeObserver.queue = queue

        def cb_finished(session: BalderSession):
            try:
                cb_validate_finished_session(session)
            except Exception:
                traceback.print_exception(*sys.exc_info())
                exit(255)

        def cb_balder_exc(exception: BalderException):
            try:
                cb_balder_exception(exception)
            except Exception:
                traceback.print_exception(*sys.exc_info())
                exit(255)

        print("\n", flush=True)
        _console_balder_debug(cmd_args=args, working_dir=env_dir, cb_session_created=set_queue,
                              cb_run_finished=cb_finished, cb_balder_exc=cb_balder_exc)
        print("\n")

    @staticmethod
    @abstractmethod
    def validate_finished_session(session: BalderSession):
        """
        This method can be used to validate the session after the process execution was finished. This method will be
        executed in separate test process!

        NOTE: All exception that are thrown inside this callback will result in status code 255

        :param session: the balder session object
        """

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        """
        This callback can be used to handle an unexpected error

        :param exc: the exception that was thrown
        """
        pass



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_scenario_check_no_reduction_of_diff_vdevice_cnns(balder_working_dir):
    """
    This testcase executes an expanded version of the CALCULATOR environment.

    It defines three new features ``VDeviceHelperFeature1``, ``VDeviceHelperFeature2`` and ``VDeviceHelperFeature3``
    that are all used by the ``ScenarioAdding.NumberOneDevice`` and the ``SetupPythonAdd.NumberProvider1``.
    Every of these features has a VDevice that reduces the connection tree:

    * ``VDeviceHelperFeature1``: ``SimulatedChildConnection.based_on(SimulatedParentConnection)``
    * ``VDeviceHelperFeature2``: ``SimulatedChildConnection``
    * ``VDeviceHelperFeature3``: ``MySimplySharedMemoryConnection``

    The VDevices are mapped on scenario level here!

    The devices (both scenario and setup) have a connection
    ``SimulatedChildConnection.based_on(SimulatedParentConnection) AND MySimplySharedMemoryConnection``, that should
    work with the different mapped VDevice requirements of these features.

    The test should run successfully with exactly one variation (these features are only provided for one ``Number*``
    device).
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.setup_executors) == 1, \
        "not exactly one setup executor found"
    assert session.executor_tree.setup_executors[0].base_setup_class.__class__.__name__ == \
           "SetupPythonAdd", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors) == 1, \
        "not exactly one scenario executor found"
    assert session.executor_tree.setup_executors[0].scenario_executors[0].base_scenario_class.__class__.__name__ == \
           "ScenarioAdding", "wrong scenario class was executed"
    assert len(session.executor_tree.setup_executors[0].scenario_executors[0].variation_executors) == 1, \
        "not exactly two variation executor found"



from ..lib.features import AddCalculateFeature, ProvidesANumberFeature
from ..lib.utils import SharedObj


class PyAddProvideANumber(ProvidesANumberFeature):

    def get_number(self):
        return self.the_number

    def sends_the_number(self):
        SharedObj.shared_mem_list.append(self.the_number)


class PyAddCalculate(AddCalculateFeature):

    def get_numbers(self):
        self.all_numbers = SharedObj.shared_mem_list

    def add_numbers(self):
        result = 0
        for cur_number in self.all_numbers:
            result += cur_number
        return result



import balder


class MySimplySharedMemoryConnection(balder.Connection):
    """This is a simply shared memory connection"""
    pass


@balder.insert_into_tree()
class SimulatedParentConnection(balder.Connection):
    pass


@balder.insert_into_tree(parents=[SimulatedParentConnection])
class SimulatedChildConnection(balder.Connection):
    pass



import balder
from ...lib.features import FeatureOfRelevanceLvl1, FeatureI, FeatureII
from ...balderglob import RuntimeObserver


class FeatureOfRelevanceLvl2(FeatureOfRelevanceLvl1):
    """this feature will be used nowhere"""

    class VDeviceWithI(balder.VDevice):
        """is newly defined here"""
        i = FeatureI()

    class VDeviceWithII(FeatureOfRelevanceLvl1.VDeviceWithII):
        i = FeatureII()

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureOfRelevanceLvl2, FeatureOfRelevanceLvl2.do_something,
            "enter `FeatureOfRelevanceLvl2.do_something`", category="feature")



from __future__ import annotations
from typing import Type, Union

import inspect
from _balder.scenario import Scenario
from _balder.utils import get_class_that_defines_method


def covered_by(item: Union[Type[Scenario], callable, None]):
    """
    This decorator defines that there exists another Scenario class or test method item that has a similar
    implementation like the decorated :class:`Scenario` class or the decorated test method.

    :param item: the element which contains the whole implementation of this scenario or test method
    """

    if item is None:
        pass
    elif callable(item) and inspect.isfunction(item) and item.__name__.startswith("test_") and \
            issubclass(get_class_that_defines_method(item), Scenario):
        pass
    elif isinstance(item, type) and issubclass(item, Scenario):
        pass
    else:
        raise TypeError("the given element for `item` must be a `Scenario` (or a subclass thereof) or a test method of "
                        "a scenario class (has to start with `test_`)")

    class CoveredByDecorator:
        """decorator class for `@covered_by` decorator"""
        def __init__(self, func):
            self.func = func

            if inspect.isclass(func):
                # it must be a class decorator
                if not issubclass(func, Scenario):
                    raise TypeError(f"The decorator `@covered_by` may only be used for `Scenario` objects or for test "
                                    f"methods of one `Scenario` object. This is not possible for the applied class "
                                    f"`{func.__name__}`.")
                if not hasattr(func, '_covered_by'):
                    func._covered_by = {}
                if func not in func._covered_by.keys():
                    func._covered_by[func] = []
                if item is None:
                    # reset it
                    func._covered_by[func] = []
                elif item not in func._covered_by[func]:
                    func._covered_by[func].append(item)
            elif inspect.isfunction(func):
                # work will done in `__set_name__`
                pass
            else:
                raise TypeError(f"The use of the `@covered_by` decorator is not allowed for the `{str(func)}` element. "
                                f"You should only use this decorator for `Scenario` elements or test method elements "
                                f"of a `Scenario` object")

        def __set_name__(self, owner, name):
            if issubclass(owner, Scenario):
                if not inspect.isfunction(self.func):
                    raise TypeError("the use of the `@covered_by` decorator is only allowed for `Scenario` objects and "
                                    "test methods of `Scenario` objects")
                if not name.startswith('test_'):
                    raise TypeError(f"the use of the `@covered_by` decorator is only allowed for `Scenario` objects "
                                    f"and test methods of `Scenario` objects - the method `{owner.__name__}.{name}` "
                                    f"does not start with `test_` and is not a valid test method")

                if not hasattr(owner, '_covered_by'):
                    owner._covered_by = {}
                if self.func not in owner._covered_by.keys():
                    owner._covered_by[self.func] = []
                if item is None:
                    # reset it
                    owner._covered_by[self.func] = []
                elif item not in owner._covered_by[self.func]:
                    owner._covered_by[self.func].append(item)
            else:
                raise TypeError(f"The use of the `@covered_by` decorator is not allowed for methods of a "
                                f"`{owner.__name__}`. You should only use this decorator for `Scenario` objects or "
                                f"valid test methods of a `Scenario` object")

            setattr(owner, name, self.func)

    return CoveredByDecorator



import balder
import balder.connections as cnns
from .features_setup import PyAddCalculate, PyAddProvideANumber


class SetupPythonAdd(balder.Setup):

    class Calculator(balder.Device):
        calc = PyAddCalculate()

    @balder.connect(Calculator, over_connection=cnns.HttpConnection)
    class NumberProvider1(balder.Device):
        n = PyAddProvideANumber()

    class NumberProvider2(balder.Device):
        n = PyAddProvideANumber()




from _balder.balder_session import BalderSession
from _balder.exceptions import FeatureOverwritingError, BalderException

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0FeatOverwriteIllegallyBefsce(Base0EnvtesterClass):
    """
    This testcase is a modified version of the basic ENVTESTER environment. It is limited to the `ScenarioA` and the
    `SetupA` and uses a feature `FeatureOfRelevance` for testing the correct overwriting of VDevices BEFORE the feature
    is being instantiated in a SCENARIO.

    This test defines a new feature `FeatureOfRelevanceLvl1` with two vDevices. This feature will be overwritten by the
    direct child `FeatureOfRelevanceLvl2`, that will be used in our scenario and also be mapped to a vDevice.
    Of course the setup device also implements a child class (`SetupFeatureOfRelevanceLvl3`) of it.

    It tries to overwrite a feature of the active vDevice from `FeatureOfRelevanceLvl2` with a class that is not allowed
    here (no direct child of overwritten one).

    The test should terminate successful and no error should occur!
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self) -> int:
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, FeatureOverwritingError), 'unexpected error type'
        assert exc.args[0] == "you are trying to overwrite an existing vDevice Feature property `i` in vDevice " \
                              "`FeatureOfRelevanceLvl2.VDeviceWithI` from the parent vDevice class " \
                              "`FeatureOfRelevanceLvl1.VDeviceWithI` - this is only possible with a child (or with " \
                              "the same) feature class the parent uses (in this case the `IllegalFeature`)"

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree is None, "test session does not terminates before collector work was done"



from ...lib.features import NewlyDefinedFeature
from ...balderglob import RuntimeObserver


class NewlySetupFeature(NewlyDefinedFeature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, NewlySetupFeature, NewlySetupFeature.do_something, "enter `NewlySetupFeature.do_something`",
            category="feature")



import balder
from ..lib.utils import SharedObj


class SetupPythonAdd(balder.Setup):

    # NOTE: This setup does not implement some devices

    @balder.fixture(level="testcase")
    def cleanup_memory(self):
        SharedObj.shared_mem_list = []



import logging
import balder
from balder.connections import UsbConnection, WifiConnection, IPv4Connection, EthernetConnection, \
    OpticalFiberConnection, TcpIPv4Connection, UdpIPv4Connection, WirelessLanConnection
from balder import Connection
from balder.exceptions import IllegalConnectionTypeError

logger = logging.getLogger(__name__)


def test_based_on_with_correct_single_resolved_members():
    """
    This test creates a simple connection tree that should be possible. This tree is already resolved and single!
    """
    subtree = IPv4Connection.based_on(EthernetConnection)
    logger.info("creation of a subtree with direct parent member is working as expected")
    assert subtree.is_resolved(), "the single resolved created subtree returns False while calling method " \
                                  "`is_resolved()`"
    assert subtree.is_single(), "the single resolved created subtree returns False while calling method `is_single()`"


def test_based_on_with_correct_non_resolved_members():
    """
    This test creates a simple connection tree that isn't resolved. Then it checks if this subtree matches the same
    subtree with all resolved members. In addition, the test secures that the subtree does not match with an inner
    subtree of itself.
    """
    subtree = IPv4Connection.based_on(OpticalFiberConnection)
    assert not subtree.is_resolved(), "the non-resolved created subtree returns True while calling method " \
                                      "`is_resolved()`"
    assert not subtree.is_single(), "the non-resolved created subtree returns True while calling method `is_single()`"
    assert subtree == IPv4Connection.based_on(EthernetConnection.based_on(OpticalFiberConnection)), \
        "comparing with complete created subtree returned true"
    assert subtree != IPv4Connection.based_on(EthernetConnection), \
        "comparing with a subtree of created subtree returned false"


def test_based_on_with_wrong_members():
    """
    This test checks if the `based_on` method throws the correct exception after the statement is not allowed.
    """
    try:
        UsbConnection.based_on(WifiConnection)
        assert False, "system thinks that USB is based on WIFI connection - and allows the creation of this tree"
    except IllegalConnectionTypeError as exc:
        logger.info(f"system raises error as expected (error message: `{str(exc)}`)")


def test_check_if_connection_container_is_resolved():
    """
    This test creates a simple connection tree that uses the raw `balder.Connection` as a container. The tree is
    already resolved but not single! The test secures that balder validates this correctly for all items of the
    container.
    """
    subtree = Connection.based_on(IPv4Connection.based_on(EthernetConnection), UdpIPv4Connection)
    logger.info("creation of a subtree with a container class and two nested other connections is working as expected")
    assert subtree.is_resolved(), "the single resolved created subtree returns False while calling method " \
                                  "`is_resolved()`"
    assert not subtree.is_single(), "the single resolved created subtree returns True while calling method " \
                                    "`is_single()`"


def test_check_contained_in():
    """
    This test checks if the `contained_in()` method returns the correct values
    """
    subtree_outer = IPv4Connection.based_on(OpticalFiberConnection)
    assert Connection().contained_in(subtree_outer), 'the basic connection object `Connection` is not contained_in ' \
                                                     'the tree - that should return True'
    assert IPv4Connection().contained_in(subtree_outer), "IPv4Connection is not contained_in the tree - that should " \
                                                         "return True"
    assert EthernetConnection().contained_in(subtree_outer), "EthernetConnection is not contained_in the tree - that " \
                                                             "should return True"
    assert OpticalFiberConnection().contained_in(subtree_outer), "OpticalFiberConnection is not contained_in the tree" \
                                                                 " - that should return True"
    assert IPv4Connection.based_on(EthernetConnection).contained_in(subtree_outer), \
        "subtree with one element less is not contained_in the tree - that should return True"
    assert EthernetConnection.based_on(OpticalFiberConnection).contained_in(subtree_outer), \
        "subtree with one element less is not contained_in the tree - that should return True"
    assert subtree_outer.contained_in(subtree_outer), "contained_in with same object does not work (should return " \
                                                      "True)"
    assert IPv4Connection.based_on(OpticalFiberConnection).contained_in(subtree_outer), \
        "contained_in with same constructed object does not working (should return True)"
    assert not TcpIPv4Connection.based_on(OpticalFiberConnection).contained_in(subtree_outer), \
        "contained_in where bigger tree contained in smaller tree does not return False!"

    container_inner = Connection.based_on(UdpIPv4Connection, TcpIPv4Connection)

    assert container_inner.contained_in(UdpIPv4Connection()), \
        'the connection sub-tree with a container connection is not contained_in the outer tree ' \
        '`UdpIPv4Connection()` - that should return True'

    assert container_inner.contained_in(TcpIPv4Connection()), \
        'the connection sub-tree with a container connection is not contained_in the outer tree ' \
        '`TcpIPv4Connection()` - that should return True'

    double_parents_inner = IPv4Connection.based_on(EthernetConnection, WirelessLanConnection)

    assert double_parents_inner.contained_in(IPv4Connection.based_on(EthernetConnection)), \
        'the connection sub-tree with double parents is not contained_in the outer tree ' \
        '`EthernetConnection()` - that should return True'

    assert double_parents_inner.contained_in(IPv4Connection.based_on(WirelessLanConnection)), \
        'the connection sub-tree with double parents is not contained_in the outer tree ' \
        '`WirelessLanConnection()` - that should return True'

    and_items = Connection.based_on((EthernetConnection, WirelessLanConnection, UsbConnection))
    or_items = Connection.based_on(EthernetConnection, WirelessLanConnection, UsbConnection)

    assert not Connection.based_on((EthernetConnection, WirelessLanConnection)).contained_in(or_items)
    assert Connection.based_on((EthernetConnection, WirelessLanConnection)).contained_in(and_items)

    assert Connection.based_on(EthernetConnection).contained_in(and_items)
    assert EthernetConnection().contained_in(and_items)


def test_check_intersection_with():
    """
    This test checks if the `intersection_with()` method returns the correct values
    """
    # nonsense (should be always empty)
    assert EthernetConnection().intersection_with(IPv4Connection()) is None
    assert Connection.based_on(EthernetConnection).intersection_with(IPv4Connection()) is None
    assert EthernetConnection().intersection_with(Connection.based_on(IPv4Connection)) is None

    # empty connection container
    assert Connection().intersection_with(EthernetConnection) == EthernetConnection()
    assert EthernetConnection().intersection_with(Connection()) == EthernetConnection()

    double_parents = IPv4Connection.based_on(EthernetConnection, WirelessLanConnection)

    # simple first based on element (without container)
    assert double_parents.intersection_with(EthernetConnection()) == EthernetConnection()
    assert EthernetConnection().intersection_with(double_parents) == EthernetConnection()

    # simple second based on element (without container)
    assert double_parents.intersection_with(WirelessLanConnection()) == WirelessLanConnection()
    assert WirelessLanConnection().intersection_with(double_parents) == WirelessLanConnection()

    # simple child element (without container)
    assert double_parents.intersection_with(IPv4Connection()) == IPv4Connection()
    assert IPv4Connection().intersection_with(double_parents) == IPv4Connection()

    # simple first based on element (with container)
    assert Connection.based_on(double_parents).intersection_with(EthernetConnection()) == EthernetConnection()
    assert double_parents.intersection_with(Connection.based_on(EthernetConnection)) == EthernetConnection()
    assert Connection.based_on(EthernetConnection).intersection_with(double_parents) == EthernetConnection()
    assert EthernetConnection().intersection_with(Connection.based_on(double_parents)) == EthernetConnection()

    # simple second based on element (with container)
    assert Connection.based_on(double_parents).intersection_with(WirelessLanConnection()) == WirelessLanConnection()
    assert double_parents.intersection_with(Connection.based_on(WirelessLanConnection)) == WirelessLanConnection()
    assert Connection.based_on(WirelessLanConnection).intersection_with(double_parents) == WirelessLanConnection()
    assert WirelessLanConnection().intersection_with(Connection.based_on(double_parents)) == WirelessLanConnection()

    # simple child element (with container)
    assert Connection.based_on(double_parents).intersection_with(IPv4Connection()) == IPv4Connection()
    assert double_parents.intersection_with(Connection.based_on(IPv4Connection)) == IPv4Connection()
    assert Connection.based_on(IPv4Connection).intersection_with(double_parents) == IPv4Connection()
    assert IPv4Connection().intersection_with(Connection.based_on(double_parents)) == IPv4Connection()

    first_branch = IPv4Connection.based_on(EthernetConnection)
    second_branch = IPv4Connection.based_on(WirelessLanConnection)

    # chain first based on element (without container)
    assert double_parents.intersection_with(first_branch) == first_branch
    assert first_branch.intersection_with(double_parents) == first_branch

    # chain second based on element (without container)
    assert double_parents.intersection_with(second_branch) == second_branch
    assert second_branch.intersection_with(double_parents) == second_branch

    # chain first based on element (with container)
    assert Connection.based_on(double_parents).intersection_with(first_branch) == first_branch
    assert double_parents.intersection_with(Connection.based_on(first_branch)) == first_branch
    assert Connection.based_on(first_branch).intersection_with(double_parents) == first_branch
    assert first_branch.intersection_with(Connection.based_on(double_parents)) == first_branch

    # chain second based on element (with container)
    assert Connection.based_on(double_parents).intersection_with(second_branch) == second_branch
    assert double_parents.intersection_with(Connection.based_on(second_branch)) == second_branch
    assert Connection.based_on(second_branch).intersection_with(double_parents) == second_branch
    assert second_branch.intersection_with(Connection.based_on(double_parents)) == second_branch

    # multiple container elements with single one
    assert Connection.based_on(EthernetConnection, WirelessLanConnection).intersection_with(EthernetConnection) == \
           EthernetConnection()

    # basic AND connection tests
    assert Connection.based_on((EthernetConnection, WirelessLanConnection)).intersection_with(
        Connection.based_on((EthernetConnection, WirelessLanConnection))) == Connection.based_on(
        (EthernetConnection(), WirelessLanConnection()))
    assert Connection.based_on(EthernetConnection, (EthernetConnection, WirelessLanConnection)).intersection_with(
        Connection.based_on((EthernetConnection, WirelessLanConnection))) == Connection.based_on(
        (EthernetConnection(), WirelessLanConnection()))
    assert Connection.based_on(EthernetConnection, (EthernetConnection, WirelessLanConnection)).intersection_with(
        EthernetConnection) == EthernetConnection()



import balder
from .between_meth_var_feature import BetweenMethVarFeature
from ...balderglob import RuntimeObserver


@balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
class SetupMethVarFeature(BetweenMethVarFeature):
    """This is the feature class that is directly being used in the setup."""
    class VDeviceI(BetweenMethVarFeature.VDeviceI):
        pass

    class VDeviceII(BetweenMethVarFeature.VDeviceII):
        pass

    class VDeviceIII(BetweenMethVarFeature.VDeviceIII):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something,
            "enter `SetupMethVarFeature.do_something`",
            category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
    def do_something_as_var(self, a_int_val, a_str_val):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            f"MethodVariation<VDeviceI><SetupMethVarFeature.do_something_as_var> a_int_val={a_int_val} "
            f"a_str_val={a_str_val}", category="feature")

    @balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
    def do_something_as_var(self, a_int_val, a_str_val):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            f"MethodVariation<VDeviceII><SetupMethVarFeature.do_something_as_var> a_int_val={a_int_val} "
            f"a_str_val={a_str_val}", category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
    def do_something_as_var(self, a_int_val, a_str_val):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, SetupMethVarFeature, SetupMethVarFeature.do_something_as_var,
            f"MethodVariation<VDeviceIII><SetupMethVarFeature.do_something_as_var> a_int_val={a_int_val} "
            f"a_str_val={a_str_val}", category="feature")



import balder
import logging
from ..lib.features import FeatureVDeviceI, FeatureVDeviceII, FeatureVDeviceIII
from .setup_features import SetupMethVarFeature
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupIII(balder.Setup):
    """This is a setup that is used with `VDeviceIII` (matches with scenario A)"""

    class SetupDevice1(balder.Device):
        s_i = FeatureVDeviceI()
        s_ii = FeatureVDeviceII()
        s_iii = FeatureVDeviceIII()

    @balder.connect(SetupDevice1, over_connection=balder.Connection())
    class SetupDevice2(balder.Device):
        s_ii = SetupMethVarFeature(VDeviceIII="SetupDevice1")

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_session, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_scenario, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_variation, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

        yield

        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_testcase, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

        yield

        RuntimeObserver.add_entry(__file__, SetupIII, SetupIII.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree

from _balder.objects.connections import osi_3_network


@insert_into_tree(parents=[osi_3_network.IPv4Connection])
class TcpIPv4Connection(Connection):
    """
    Balder TCP IPv4 connection (OSI LAYER 4)
    """


@insert_into_tree(parents=[osi_3_network.IPv6Connection])
class TcpIPv6Connection(Connection):
    """
    Balder TCP IPv6 connection (OSI LAYER 4)
    """


TcpConnection = Connection.based_on(TcpIPv4Connection, TcpIPv6Connection)


@insert_into_tree(parents=[osi_3_network.IPv4Connection])
class UdpIPv4Connection(Connection):
    """
    Balder UDP IPv4 connection (OSI LAYER 4)
    """


@insert_into_tree(parents=[osi_3_network.IPv6Connection])
class UdpIPv6Connection(Connection):
    """
    Balder UDP IPv6 connection (OSI LAYER 4)
    """


UdpConnection = Connection.based_on(UdpIPv4Connection, UdpIPv6Connection)



from __future__ import annotations
from typing import Type, Union, List, Tuple

from _balder.connection import Connection


def insert_into_tree(parents: List[Union[Type[Connection], Tuple[Type[Connection]]]] = None, tree_name: str = ""):
    """
    This decorator inserts a :meth:`Connection` object into the global connection tree.

    :param parents: all items that are parents of the decorated connection

    :param tree_name: the tree name the connection should be inserted in
    """
    if parents is None:
        parents = []
    if not isinstance(parents, list):
        raise ValueError("the value of `parents` must be a `list`")

    idx = 0
    for cur_parent in parents:
        if isinstance(cur_parent, tuple):
            tuple_idx = 0
            for cur_tuple_element in cur_parent:
                if not isinstance(cur_tuple_element, type) or not issubclass(cur_tuple_element, Connection):
                    raise TypeError(
                        f"the tuple element on index {tuple_idx} (located on index {idx} of `parents`)  must be a "
                        f"`Connection` type (given element is: {str(cur_parent)})")
                tuple_idx += 1
        else:
            if not isinstance(cur_parent, type) or not issubclass(cur_parent, Connection):
                raise TypeError(
                    f"the element on index {idx} in `parents` must be a `Connection` type (given: {str(cur_parent)})")
        idx += 1

    class MyDecorator:
        """decorator class for `@insert_into_tree` decorator"""
        def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument

            nonlocal parents

            decorated_cls = args[0]

            conn_parents = decorated_cls.get_parents(tree_name=tree_name)
            conn_parents = [] if conn_parents is None else conn_parents

            conn_parents += [new_parent for new_parent in parents if new_parent not in conn_parents]
            decorated_cls.set_parents(conn_parents, tree_name=tree_name)

            return decorated_cls

    return MyDecorator



import balder


class ParentConnection(balder.Connection):
    pass


@balder.insert_into_tree(parents=[ParentConnection])
class BetweenConnection(balder.Connection):
    pass


@balder.insert_into_tree(parents=[BetweenConnection])
class ChildConnection(balder.Connection):
    pass



import balder
import logging
from ..lib.features import FeatureIII, FeatureIV
from ..lib.connections import BetweenConnection

logger = logging.getLogger(__name__)


class ScenarioB(balder.Scenario):
    """This is the scenario of category B"""

    class ScenarioDevice3(balder.Device):
        iii = FeatureIII()

    @balder.connect(ScenarioDevice3, over_connection=BetweenConnection)
    class ScenarioDevice4(balder.Device):
        iv = FeatureIV()

    def test_b_1(self):
        logger.info(f"execute testcase `{ScenarioB.test_b_1.__qualname__}`")
        assert True

    def test_b_2(self):
        logger.info(f"execute testcase `{ScenarioB.test_b_2.__qualname__}`")
        assert True



from __future__ import annotations
from typing import List, Type, Tuple, Union, Literal

import sys
import inspect
from _balder.scenario import Scenario
from _balder.exceptions import InheritanceError

MethodLiteralType = Literal["function", "classmethod", "staticmethod", "instancemethod"]


def get_scenario_inheritance_list_of(scenario: Type[Scenario]) -> List[Type[Scenario]]:
    """
    This method returns a list, beginning with the given `scenario`, of all parent classes that are from type
    :class:´Scenario`. The last element of the list is always the original :class:`Scenario` class.

    :param scenario: the scenario the method should look for in the inheritance list

    :returns: a list with all classes that are from type scenario
    """
    if scenario == Scenario:
        return [Scenario]
    base_class_of_interest = None
    for cur_parent in scenario.__bases__:
        if issubclass(cur_parent, Scenario):
            if base_class_of_interest is not None:
                raise InheritanceError("it is not allowed to have multiple parent class that inherit from the "
                                       "class `Scenario`")
            base_class_of_interest = cur_parent

    return [scenario] + get_scenario_inheritance_list_of(base_class_of_interest)


def get_class_that_defines_method(meth):
    """credits to https://stackoverflow.com/a/25959545"""
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth), meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return None  # not required since None would have been implicitly returned anyway


def inspect_method(func) -> Tuple[Union[type, None], MethodLiteralType]:
    """
    This helper function returns the related class and the type of the method (`staticmethod`, `classmethod`,
    `instancemethod` or `function`) as tuple.
    """

    def get_func_class():
        """helper method to determine the class of the `func`"""
        classes = []
        cur_frame = sys.modules[func.__module__]
        names = func.__qualname__.split('.')[:-1]
        for cur_name in names:
            cur_frame = getattr(cur_frame, cur_name)
            if isinstance(cur_frame, type):
                classes.append(cur_frame)
        return classes[-1] if classes else None

    if func.__name__ == func.__qualname__ or '.<locals>' in func.__qualname__:
        return None, 'function'
    cls = get_func_class()
    fn_type = cls.__dict__.get(func.__name__)

    if isinstance(fn_type, classmethod):
        return cls, 'classmethod'

    if isinstance(fn_type, staticmethod):
        return cls, 'staticmethod'

    if fn_type is not None and fn_type.__class__.__name__ == 'function':
        return cls, 'instancemethod'

    raise TypeError(f'unknown element type `{func}`')





class FixtureReturn:
    """helper const class for return values"""
    BALDERGLOB_SESSION = "balderglob_session_fixt"
    BALDERGLOB_SETUP = "balderglob_setup_fixt"
    BALDERGLOB_SCENARIO = "balderglob_scenario_fixt"
    BALDERGLOB_VARIATION = "balderglob_variation_fixt"
    BALDERGLOB_TESTCASE = "balderglob_testcase_fixt"

    SETUP_A_SESSION = "setup_a_session_fixt"
    SETUP_A_SETUP = "setup_a_setup_fixt"
    SETUP_A_SCENARIO = "setup_a_scenario_fixt"
    SETUP_A_VARIATION = "setup_a_variation_fixt"
    SETUP_A_TESTCASE = "setup_a_testcase_fixt"

    SETUP_B_SESSION = "setup_b_session_fixt"
    SETUP_B_SETUP = "setup_b_setup_fixt"
    SETUP_B_SCENARIO = "setup_b_scenario_fixt"
    SETUP_B_VARIATION = "setup_b_variation_fixt"
    SETUP_B_TESTCASE = "setup_b_testcase_fixt"

    SCENARIO_A_SESSION = "scenario_a_session_fixt"
    SCENARIO_A_SETUP = "scenario_a_setup_fixt"
    SCENARIO_A_SCENARIO = "scenario_a_scenario_fixt"
    SCENARIO_A_VARIATION = "scenario_a_variation_fixt"
    SCENARIO_A_TESTCASE = "scenario_a_testcase_fixt"

    SCENARIO_B_SESSION = "scenario_b_session_fixt"
    SCENARIO_B_SETUP = "scenario_b_setup_fixt"
    SCENARIO_B_SCENARIO = "scenario_b_scenario_fixt"
    SCENARIO_B_VARIATION = "scenario_b_variation_fixt"
    SCENARIO_B_TESTCASE = "scenario_b_testcase_fixt"



import balder
from ..balderglob import RuntimeObserver


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")


class FeatureII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")


class FeatureIII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIII, FeatureIII.do_something, "enter `FeatureIII.do_something`", category="feature")


class FeatureIV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureIV, FeatureIV.do_something, "enter `FeatureIV.do_something`", category="feature")


class FeatureV(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureV, FeatureV.do_something, "enter `FeatureV.do_something`", category="feature")



from ...lib.features import FeatureII
from ...balderglob import RuntimeObserver


class SetupFeatureII(FeatureII):

    class VDeviceFeatureI(FeatureII.VDeviceFeatureI):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureII, SetupFeatureII.do_something, "enter `SetupFeatureII.do_something`",
            category="feature")




import balder
from ..balderglob import RuntimeObserver
from ..lib.connections import ChildAConnection, ChildBConnection


class FeatureI(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something, "enter `FeatureI.do_something`", category="feature")

    def do_something_special(self):
        RuntimeObserver.add_entry(
            __file__, FeatureI, FeatureI.do_something_special, "enter `FeatureI.do_something_special`",
            category="feature")


class FeatureII(balder.Feature):

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something, "enter `FeatureII.do_something`", category="feature")

    def do_something_special_with_the_vdevice(self):
        RuntimeObserver.add_entry(
            __file__, FeatureII, FeatureII.do_something_special_with_the_vdevice,
            "enter `FeatureII.do_something_special_with_the_vdevice`", category="feature")



from __future__ import annotations

from _balder.connection import Connection
from _balder.decorator_insert_into_tree import insert_into_tree

from _balder.objects.connections import osi_2_datalink


@insert_into_tree(parents=[osi_2_datalink.EthernetConnection, osi_2_datalink.WirelessLanConnection])
class IPv4Connection(Connection):
    """
    Balder IPv4 connection (OSI LAYER 3)
    """


@insert_into_tree(parents=[osi_2_datalink.EthernetConnection, osi_2_datalink.WirelessLanConnection])
class IPv6Connection(Connection):
    """
    Balder IPv6 connection (OSI LAYER 3)
    """


IPConnection = Connection.based_on(IPv4Connection, IPv6Connection)


@insert_into_tree(parents=[IPv4Connection, IPv6Connection])
class IpSecConnection(Connection):
    """
    Balder IPSec connection (OSI LAYER 3)
    """


@insert_into_tree(parents=[IPv4Connection])
class ICMPv4Connection(Connection):
    """
    Balder ICMP IPv4 connection (OSI LAYER 3)
    """


@insert_into_tree(parents=[IPv6Connection])
class ICMPv6Connection(Connection):
    """
    Balder ICMP IPv6 connection (OSI LAYER 3)
    """


ICMPConnection = Connection.based_on(ICMPv4Connection, ICMPv6Connection)



from __future__ import annotations

import pathlib
import sys
import traceback
from typing import Callable, Optional, Union, List
from _balder.exit_code import ExitCode
from _balder.testresult import ResultState
from _balder.exceptions import BalderException
from _balder.balder_session import BalderSession


def console_balder(cmd_args: Optional[List[str]] = None, working_dir: Union[str, pathlib.Path, None] = None):
    """script that executes a balder session"""
    _console_balder_debug(cmd_args=cmd_args, working_dir=working_dir)


def _console_balder_debug(cmd_args: Optional[List[str]] = None, working_dir: Union[str, pathlib.Path, None] = None,
                          cb_session_created: Optional[Callable] = None, cb_run_finished: Optional[Callable] = None,
                          cb_balder_exc: Optional[Callable] = None, cb_unexpected_exc: Optional[Callable] = None):
    """helper balder execution that allows more debug access"""
    try:
        balder_session = BalderSession(cmd_args=cmd_args, working_dir=working_dir)

        if cb_session_created:
            cb_session_created(balder_session)

        balder_session.run()

        if cb_run_finished:
            cb_run_finished(balder_session)

        if balder_session.executor_tree is None:
            sys.exit(ExitCode.SUCCESS.value)
        elif balder_session.executor_tree.executor_result == ResultState.SUCCESS:
            sys.exit(ExitCode.SUCCESS.value)
        elif balder_session.executor_tree.executor_result in [ResultState.ERROR, ResultState.FAILURE]:
            # check if a BalderException was thrown too -> would be a balder environment error -> special exit code
            balder_exceptions = [cur_exc for cur_exc in balder_session.executor_tree.get_all_recognized_exception()
                                 if isinstance(cur_exc, BalderException)]
            if len(balder_exceptions) > 0:
                sys.exit(ExitCode.BALDER_USAGE_ERROR.value)
            else:
                sys.exit(ExitCode.TESTS_FAILED.value)

    except BalderException as exc:
        # a balder usage error occurs
        if cb_balder_exc:
            cb_balder_exc(exc)
        traceback.print_exception(*sys.exc_info())
        sys.exit(ExitCode.BALDER_USAGE_ERROR.value)
    except Exception as exc:
        # a unexpected error occurs
        if cb_unexpected_exc:
            cb_unexpected_exc(exc)
        traceback.print_exception(*sys.exc_info())
        sys.exit(ExitCode.UNEXPECTED_ERROR.value)



from .feature_i import SetupFeatureI
from .feature_ii import SetupFeatureII
from .feature_iii import SetupFeatureIII
from .feature_iv import SetupFeatureIV
from .feature_v import SetupFeatureV

__all__ = [
    "SetupFeatureI",
    "SetupFeatureII",
    "SetupFeatureIII",
    "SetupFeatureIV",
    "SetupFeatureV"
]



import logging
import balder
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature
from balder.exceptions import AccessToUnmappedVDeviceException


logger = logging.getLogger(__name__)


class ScenarioAddingNoMapping(balder.Scenario):

    class Calculator(balder.Device):
        adds = AddCalculateFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()

    def test_add_two_numbers(self):
        self.NumberOneDevice.number.set_number(3)
        self.NumberTwoDevice.number.set_number(4)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 7, "result is not as expected"



from _balder.testresult import ResultState
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0TreecheckFixtSetupaSetupTeardown(Base0EnvtesterClass):
    """
    This testcase executes the basic envtester and forces an error on a specific given position. The test checks if the
    system behaviour is as expected.

    file: ``setups/setup_a.py``
    class: ``SetupA``
    method/function: ``fixture_setup``
    part: ``teardown``
    """

    @property
    def cmd_args(self):
        return [
            '--test-error-file', 'setups/setup_a.py',
            '--test-error-cls', 'SetupA',
            '--test-error-meth', 'fixture_setup',
            '--test-error-part', 'teardown',
        ]

    @property
    def expected_exit_code(self):
        return 1

    @property
    def expected_data(self):
        return (
            # FIXTURE-CONSTRUCTION: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "construction"},
            [
                (
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_1
                            {"cls": "ScenarioA", "meth": "test_a_1"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_a_2
                            {"cls": "ScenarioA", "meth": "test_a_2"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioA.fixture_testcase
                            {"cls": "ScenarioA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupA.fixture_testcase
                            {"cls": "SetupA", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_variation
                    {"cls": "ScenarioA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_variation
                    {"cls": "SetupA", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_scenario
                    {"cls": "ScenarioA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_scenario
                    {"cls": "SetupA", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_setup
                    {"cls": "ScenarioA", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupA.fixture_setup
                    {"cls": "SetupA", "meth": "fixture_setup", "part": "teardown"},

                    # error occurs here
                    # {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    # {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},

                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
                (
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "construction"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "construction"},
                    # FIXTURE-CONSTRUCTION: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-CONSTRUCTION: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "construction"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    [
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_1
                            {"cls": "ScenarioB", "meth": "test_b_1"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                        (
                            # FIXTURE-CONSTRUCTION: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "construction"},
                            # FIXTURE-CONSTRUCTION: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-CONSTRUCTION: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "construction"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # TESTCASE: ScenarioA.test_b_2
                            {"cls": "ScenarioB", "meth": "test_b_2"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: ScenarioB.fixture_testcase
                            {"cls": "ScenarioB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: SetupB.fixture_testcase
                            {"cls": "SetupB", "meth": "fixture_testcase", "part": "teardown"},
                            {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                            {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                            # FIXTURE-TEARDOWN: balderglob_fixture_testcase
                            {"file": "balderglob.py", "meth": "balderglob_fixture_testcase", "part": "teardown"},
                        ),
                    ],
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_variation
                    {"cls": "ScenarioB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_variation
                    {"cls": "SetupB", "meth": "fixture_variation", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_variation
                    {"file": "balderglob.py", "meth": "balderglob_fixture_variation", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_scenario
                    {"cls": "ScenarioB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_scenario
                    {"cls": "SetupB", "meth": "fixture_scenario", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_scenario
                    {"file": "balderglob.py", "meth": "balderglob_fixture_scenario", "part": "teardown"},
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_setup
                    {"cls": "ScenarioB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: SetupB.fixture_setup
                    {"cls": "SetupB", "meth": "fixture_setup", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                    # FIXTURE-TEARDOWN: balderglob_fixture_setup
                    {"file": "balderglob.py", "meth": "balderglob_fixture_setup", "part": "teardown"},
                ),
            ],
            [
                (
                    # FIXTURE-TEARDOWN: ScenarioA.fixture_session
                    {"cls": "ScenarioA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: ScenarioB.fixture_session
                    {"cls": "ScenarioB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "FeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "FeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            [
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupA", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureI", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureII", "meth": "do_something", "category": "feature"},
                ),
                (
                    # FIXTURE-TEARDOWN: SetupA.fixture_session
                    {"cls": "SetupB", "meth": "fixture_session", "part": "teardown"},
                    {"cls": "SetupFeatureIII", "meth": "do_something", "category": "feature"},
                    {"cls": "SetupFeatureIV", "meth": "do_something", "category": "feature"},
                )
            ],
            # FIXTURE-TEARDOWN: balderglob_fixture_session
            {"file": "balderglob.py", "meth": "balderglob_fixture_session", "part": "teardown"},
        )

    @staticmethod
    def validate_finished_session(session: BalderSession):

        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.ERROR, "test session does not terminates with ERROR"

        assert session.executor_tree.construct_result.result == ResultState.SUCCESS, \
            "global executor tree construct part does not set ResultState.SUCCESS"
        assert session.executor_tree.body_result.result == ResultState.ERROR, \
            "global executor tree body part does not set ResultState.ERROR"
        assert session.executor_tree.teardown_result.result == ResultState.SUCCESS, \
            "global executor tree teardown part does not set ResultState.SUCCESS"
        for cur_setup_executor in session.executor_tree.setup_executors:
            if cur_setup_executor.base_setup_class.__class__.__name__ == "SetupA":
                assert cur_setup_executor.executor_result == ResultState.ERROR, \
                    "the setup executor does not have result ERROR"

                assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
                assert cur_setup_executor.body_result.result == ResultState.SUCCESS
                assert cur_setup_executor.teardown_result.result == ResultState.ERROR

            else:
                # in this branch everything have to be successfully
                assert cur_setup_executor.executor_result == ResultState.SUCCESS, \
                    "the setup executor does not have result SUCCESS"

                assert cur_setup_executor.construct_result.result == ResultState.SUCCESS
                assert cur_setup_executor.body_result.result == ResultState.SUCCESS
                assert cur_setup_executor.teardown_result.result == ResultState.SUCCESS

            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                assert cur_scenario_executor.executor_result == ResultState.SUCCESS, \
                    "the scenario executor does not have result ERROR"

                assert cur_scenario_executor.construct_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.body_result.result == ResultState.SUCCESS
                assert cur_scenario_executor.teardown_result.result == ResultState.SUCCESS

                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    assert cur_variation_executor.executor_result == ResultState.SUCCESS, \
                        "the variation executor does not have result NOT_RUN"

                    assert cur_variation_executor.construct_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.body_result.result == ResultState.SUCCESS
                    assert cur_variation_executor.teardown_result.result == ResultState.SUCCESS

                    for cur_testcase_executor in cur_variation_executor.testcase_executors:
                        assert cur_testcase_executor.executor_result == ResultState.SUCCESS, \
                            "the testcase executor does not have result SUCCESS"

                        assert cur_testcase_executor.construct_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.body_result.result == ResultState.SUCCESS
                        assert cur_testcase_executor.teardown_result.result == ResultState.SUCCESS



from __future__ import annotations
from typing import Union, List, Type, TYPE_CHECKING

from _balder.executor.setup_executor import SetupExecutor
from _balder.executor.basic_executor import BasicExecutor
from _balder.fixture_manager import FixtureManager
from _balder.testresult import ResultState, BranchBodyResult
from _balder.previous_executor_mark import PreviousExecutorMark

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.executor.scenario_executor import ScenarioExecutor
    from _balder.executor.variation_executor import VariationExecutor
    from _balder.executor.testcase_executor import TestcaseExecutor


class ExecutorTree(BasicExecutor):
    """
    This class is the root object of the executor tree structure
    """

    def __init__(self, fixture_manager: FixtureManager):
        super().__init__()
        self._setup_executors: List[SetupExecutor] = []
        self._fixture_manager = fixture_manager

        # contains the result object for the BODY part of this branch (will be overwritten in :class:`TestcaseExecutor`)
        self.body_result = BranchBodyResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> List[BasicExecutor]:
        return self.setup_executors

    @property
    def base_instance(self) -> object:
        """returns None because this element is a ExecutorTree"""
        return None

    @property
    def setup_executors(self) -> List[SetupExecutor]:
        """returns all setup executors of this tree"""
        return self._setup_executors

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the fixture manager of this tree"""
        return self._fixture_manager

    @property
    def parent_executor(self) -> None:
        return None

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self):
        pass

    def _body_execution(self):
        for cur_setup_executor in self.setup_executors:
            if cur_setup_executor.has_runnable_elements():
                cur_setup_executor.execute()
            elif cur_setup_executor.prev_mark == PreviousExecutorMark.SKIP:
                cur_setup_executor.set_result_for_whole_branch(ResultState.SKIP)
            elif cur_setup_executor.prev_mark == PreviousExecutorMark.COVERED_BY:
                cur_setup_executor.set_result_for_whole_branch(ResultState.COVERED_BY)
            else:
                cur_setup_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self):
        pass

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_all_scenario_executors(self) -> List[ScenarioExecutor]:
        """
        returns a list with all scenario executors
        """
        all_scenario_executor = []
        for cur_setup_executor in self.setup_executors:
            all_scenario_executor += cur_setup_executor.scenario_executors
        return all_scenario_executor

    def get_all_variation_executors(self) -> List[VariationExecutor]:
        """
        returns a list with all variation executors
        """
        all_variation_executor = []
        for cur_scenario_executor in self.get_all_scenario_executors():
            all_variation_executor += cur_scenario_executor.variation_executors
        return all_variation_executor

    def get_all_testcase_executors(self) -> List[TestcaseExecutor]:
        """
        returns a list with all testcase executors
        """
        all_testcase_executor = []
        for cur_scenario_executor in self.get_all_scenario_executors():
            for cur_variation_executor in cur_scenario_executor.variation_executors:
                all_testcase_executor += cur_variation_executor.testcase_executors
        return all_testcase_executor

    def add_setup_executor(self, setup_executor: SetupExecutor):
        """
        This method adds a new :class:`SetupExecutor` to the child element's list of this tree object
        """
        if not isinstance(setup_executor, SetupExecutor):
            raise TypeError("the given object `setup_executor` must be of type `SetupExecutor`")
        if setup_executor in self._setup_executors:
            raise ValueError("the given object `setup_executor` already exists in child list")
        self._setup_executors.append(setup_executor)

    def get_executor_for_setup(self, setup: Type[Setup]) -> Union[SetupExecutor, None]:
        """
        This method searches for a SetupExecutor in the internal list for the given :class:`Setup` type

        :param setup: the setup class for which the executor is being searched for

        :return: returns the associated SetupExecutor or None if there are no matches for the given type
        """
        for cur_setup_executor in self._setup_executors:
            if cur_setup_executor.base_setup_class.__class__ == setup:
                return cur_setup_executor
        # can not find some
        return None

    def cleanup_empty_executor_branches(self):
        to_remove_executor = []
        for cur_setup_executor in self.setup_executors:
            cur_setup_executor.cleanup_empty_executor_branches()
            if len(cur_setup_executor.scenario_executors) == 0:
                # remove this whole executor because it has no children anymore
                to_remove_executor.append(cur_setup_executor)
        for cur_setup_executor in to_remove_executor:
            self._setup_executors.remove(cur_setup_executor)

    def execute(self) -> None:
        """
        This method executes this branch of the tree
        """
        start_text = "START TESTSESSION"
        end_text = "FINISH TESTSESSION"
        line_length = 120

        def print_line(text):
            full_text = int((line_length - (len(start_text) + 2)) / 2) * "=" + " " + text + " "
            full_text += "=" * (line_length - len(full_text))
            print(full_text)

        print_line(start_text)
        # check if there exists runnable elements
        runnables = [cur_exec.has_runnable_elements() for cur_exec in self.setup_executors]
        one_or_more_runnable_setups = None if len(runnables) == 0 else max(runnables)
        if one_or_more_runnable_setups:
            super().execute()
        else:
            print("NO EXECUTABLE SETUPS/SCENARIOS FOUND")
        print_line(end_text)
        summary = self.testsummary()
        is_first = True
        for cur_key, cur_val in summary.items():
            if is_first:
                is_first = False
            else:
                print(" | ", end="")
            print(f"TOTAL {cur_key.value}: {cur_val}", end="")
        print("")

    def print_tree(self) -> None:
        """this method is an auxiliary method which outputs the entire tree"""
        print("RESOLVING OVERVIEW", end="\n\n")
        for cur_setup_executor in self.setup_executors:
            for cur_scenario_executor in cur_setup_executor.scenario_executors:
                for cur_variation_executor in cur_scenario_executor.variation_executors:
                    print(f"Scenario `{cur_scenario_executor.base_scenario_class.__class__.__qualname__}` <-> "
                          f"Setup `{cur_setup_executor.base_setup_class.__class__.__qualname__}`")
                    mapping_printings = {}
                    for cur_scenario_device, cur_setup_device in cur_variation_executor.base_device_mapping.items():
                        mapping_printings[f"   {cur_scenario_device.__qualname__}"] = str(cur_setup_device.__qualname__)
                    max_len = max(len(cur_elem) for cur_elem in mapping_printings.keys())
                    for cur_key, cur_val in mapping_printings.items():
                        print(("{:<" + str(max_len) + "} = {}").format(cur_key, cur_val))
                    for cur_testcase_excutor in cur_variation_executor.testcase_executors:
                        print(f"   -> Testcase<{cur_testcase_excutor.base_testcase_callable.__qualname__}>")
